-----------------------------------
-- Setting up scope and libs
-----------------------------------

local AddonName = select(1, ...);
iGear = LibStub("AceAddon-3.0"):NewAddon(AddonName, "AceEvent-3.0");

local L = LibStub("AceLocale-3.0"):GetLocale(AddonName);

local LibQTip = LibStub("LibQTip-1.0");
local LibCrayon = LibStub("LibCrayon-3.0");

local _G = _G; -- upvalueing done here since I always will call _G.func()

-----------------------------------------
-- Variables, functions and colors
-----------------------------------------

local Tooltip; -- our QTip object
local RepairCosts = 0; -- calculated repaircosts

local COLOR_GOLD = "|cfffed100%s|r";

-- This is iGears static info table which lets the addon know what to do with an itemslot.
-- In front of each index are two infos displayed. The first can be + (set by author) or - (set by iGear).
-- The second info is the data type in short, where i means int, s means string, b means bool. Simple!
--  + s  1: API internal slot name
--  - i  2: the API internal slot ID
--  + b  3: indicates whether a slot can be repaired or not
--  - i  4: the slot durability. Here the current item durability will be stored
--  - i  5: indicates how much gem slots are empty in the current slot
--  - i  6: stores the repair costs of the slot
--  + b  7: indicates whether a slot can be enchanted or not
--  - i  8: stores the enchant id of the enchant, where 0 means not enchanted
--  - b  9: indicates whether an item is equipped in the slot
--  - s 10: stores the slot color (legendary, epic, rare, common, uncommon, waste)
--  - s 11: stores the name of the weared item in this slot
local EquipSlots = {
--   1            2  3      4  5  6  7      8  9      10  11
	{"HeadSlot",		0, true,  0, 0, 0, true,  0, false, "", ""},
	{"NeckSlot",		0, false, 0, 0, 0, false, 0, false, "", ""},
	{"ShoulderSlot",	0, true,  0, 0, 0, true,  0, false, "", ""},
	{"BackSlot",		0, false, 0, 0, 0, true,  0, false, "", ""},
	{"ChestSlot",		0, true,  0, 0, 0, true,  0, false, "", ""},
	{"WristSlot",		0, true,  0, 0, 0, true,  0, false, "", ""},
	{"HandsSlot",		0, true,  0, 0, 0, true,  0, false, "", ""},
	{"WaistSlot",		0, true,  0, 0, 0, false, 0, false, "", ""},
	{"LegsSlot",		0, true,  0, 0, 0, true,  0, false, "", ""},
	{"FeetSlot",		0, true,  0, 0, 0, true,  0, false, "", ""},
	{"Finger0Slot",		0, false, 0, 0, 0, false, 0, false, "", ""},
	{"Finger1Slot",		0, false, 0, 0, 0, false, 0, false, "", ""},
	{"Trinket0Slot",	0, false, 0, 0, 0, false, 0, false, "", ""},
	{"Trinket1Slot",	0, false, 0, 0, 0, false, 0, false, "", ""},
	{"MainHandSlot",	0, true,  0, 0, 0, true,  0, false, "", ""},
	{"SecondaryHandSlot",0,true,  0, 0, 0, true,  0, false, "", ""},
	{"RangedSlot",		0, false, 0, 0, 0, false, 0, false, "", ""}
};
-- Yes, I prevent using key/value pairs in tables mostly, so every table index gets a name here:
local SLOT_NAME = 1;
local SLOT_ID = 2;
local SLOT_CAN_DURABILITY = 3;
local SLOT_DURABILITY = 4;
local SLOT_GEMS_EMPTY = 5;
local SLOT_REPAIR_COST = 6;
local SLOT_CAN_ENCHANT = 7;
local SLOT_ENCHANT = 8;
local SLOT_IS_EQUIPPED = 9;
local SLOT_COLOR = 10;
local SLOT_ITEM_NAME = 11;

-- clears a table and frees memory on the next garbage collect
local function tclear(t, wipe)
	if( type(t) ~= "table" ) then return end;
	for k in pairs(t) do
		t[k] = nil;
	end
	t[''] = 1;
	t[''] = nil;
	if( wipe ) then
		t = nil;
	end
end

--------------------------------------------------------------------
-- Initialize static table changes according to players class
--------------------------------------------------------------------

-- TO BE REMOVED WITH TOC 50000 AND ABOVE!
--=========================================
-- There are some differences between the classes. Hunter bows (ranged slot), can be enchanted.
-- Paladin Relics (ranged slot, too), canNOT be enchanted. So, when I play a Hunter, ranged SLOT_CAN_ENCHANT must be true.
do
	local _, class = _G.UnitClass("player");
	for _, slot in ipairs(EquipSlots) do
		
		-- WARRIOR can wear bows
		-- ROGUE can wear bows
		-- HUNTER can wear bows
		if( slot[SLOT_NAME] == "RangedSlot" ) then
			if( class == "WARRIOR" or class == "ROGUE" or class == "HUNTER" ) then
				slot[SLOT_CAN_ENCHANT] = true;
			end
		end
		
	end
end

-----------------------------
-- Setting up the feed
-----------------------------

iGear.Feed = LibStub("LibDataBroker-1.1"):NewDataObject(AddonName, {
	type = "data source",
	text = "",
	icon = "Interface\\Minimap\\Tracking\\Repair",
});

iGear.Feed.OnEnter = function(anchor)
	-- when there are no problems, we don't need the tooltip.
	if( iGear:GetNumProblems(1) == 0 ) then
		return;
	end

	-- force other i-Series qtip's to hide, when this tip is to be shown
	for k, v in LibQTip:IterateTooltips() do
		if( type(k) == "string" and strsub(k, 1, 6) == "iSuite" ) then
			v:Release(k);
		end
	end
	
	Tooltip = LibQTip:Acquire("iSuite"..AddonName);
	Tooltip:SetAutoHideDelay(0.1, anchor);
	Tooltip:SmartAnchorTo(anchor);
	iGear:UpdateTooltip();
	Tooltip:Show();
end

----------------------
-- OnInitialize
----------------------

function iGear:OnInitialize()
	-- on first run, we fetch the slot IDs by their slot name by some API calls. Why? Slot IDs could change, slot names doesn't ever change.
	for _, slot in ipairs(EquipSlots) do
		slot[SLOT_ID] = _G.GetInventorySlotInfo(slot[SLOT_NAME]);
	end
	
	self:RegisterEvent("PLAYER_DEAD", "DurabilityChange");
	self:RegisterEvent("PLAYER_UNGHOST", "DurabilityChange");
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "DurabilityChange");
	self:RegisterEvent("UPDATE_INVENTORY_DURABILITY", "DurabilityChange");
end

-----------------------
-- Num problems
-----------------------

function iGear:GetProblem(s, repair)
	if( not s[SLOT_IS_EQUIPPED] ) then
		return false;
	end
	
	if( repair and s[SLOT_CAN_DURABILITY] and s[SLOT_DURABILITY] ~= 100 ) then
		return true;
	elseif( s[SLOT_GEMS_EMPTY] > 0 ) then
		return true;
	elseif( s[SLOT_CAN_ENCHANT] and s[SLOT_ENCHANT] == 0 ) then
		return true;
	end
	
	return false;
end

function iGear:GetNumProblems(repair)
	local problems = 0;
	
	for _, s in ipairs(EquipSlots) do
		if( self:GetProblem(s, repair) ) then
			problems = problems + 1;
		end
	end
	
	return problems;
end

--------------------------
-- Get Item Infos
--------------------------

-- This frame is a fake GameTooltip object which allows us to scan data from it
_G.CreateFrame("GameTooltip", "iGearScanTip", UIParent, "GameTooltipTemplate");

local NumMissingGems;
do
	-- there is no API call which returns missing gems in a slot. We fetch the item tooltip and scan it for gem textures.
	-- This are the default textures for gem sockets, so they are just shown in the tip when gems are missing.
	local scanForTex = {
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-Meta",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-Red",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-Yellow",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-Blue",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-CogWheel",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-Hydraulic",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket-Prismatic",
		"Interface\\ItemSocketingFrame\\UI-EmptySocket"
	};
	
	NumMissingGems = function(slotNum)
		local missing = 0
		
		-- Textures aren't reset when fetching new tooltips. So we clear them manually out of the tip.
		for i = 1, 4 do
			if ( _G["iGearScanTipTexture"..i] ) then
				_G["iGearScanTipTexture"..i]:SetTexture("");
			end
		end
		
		_G.iGearScanTip:ClearLines(); -- Clear all text from the tip, too.
		_G.iGearScanTip:SetInventoryItem("player", slotNum); -- Fetch the tooltip of a given item slot.
		
		local tex;
		for i = 1, 4 do
			tex = _G["iGearScanTipTexture"..i]:GetTexture();
			
			if( tex ) then
				for j = 1, #scanForTex do
					if( tex == scanForTex[j] ) then
						missing = missing + 1;
					end
				end
			end
		end
		
		_G.iGearScanTip:Hide();
		return missing;
	end
end

local function GetItemEquippedAndCost(slotNum)
	_G.iGearScanTip:ClearLines();
	
	local equipped, _, cost = _G.iGearScanTip:SetInventoryItem("player", slotNum);
	
	_G.iGearScanTip:Hide();
	return equipped, cost;
end

local function GetItemDurability(s)
	-- no durability if item has no durability on it
	if( not s[SLOT_CAN_DURABILITY] ) then
		return;
	end
	
	-- if the item is broken, durability automatically is 0!
	if( _G.GetInventoryItemBroken("player", s[SLOT_ID]) ) then
		return 0;
	end
	local current, maximum = _G.GetInventoryItemDurability(s[SLOT_ID]);
	
	if( not current or not maximum ) then
		return;
	end
	-- let's do some math to get the percent item durability.
	local durability = 100 * current / maximum;
	
	if( durability - abs(durability) >= 0.5 ) then
		durability = ceil(durability);
	else
		durability = floor(durability);
	end
	
	return durability;
end

local function GetItemInfos(slotNum)
	local link = _G.GetInventoryItemLink("player", slotNum);
	local _, _, color, enchant, name = string.find(link,"|c%x%x(%x*)|Hitem:%d+:(%d+):%d+:%d+:%d+:%d+:%-?%d+:%-?%d+:%d+:%d+|h%[([^%]]*)%]");
	
	return color, enchant, name;
end

----------------------
-- EventHandler
----------------------

function iGear:DurabilityChange()
	RepairCosts = 0;

	local smallest_durability = 100; -- iGear displays the smallest durability on the feed.
	local equipped, repcosts, durability, enchant;
	
	for _, s in ipairs(EquipSlots) do
		equipped, repcosts = GetItemEquippedAndCost(s[SLOT_ID]);
		
		if( not equipped ) then
			s[SLOT_IS_EQUIPPED] = false;
		else
			s[SLOT_IS_EQUIPPED] = true;
			s[SLOT_CAN_DURABILITY] = true;
			
			-- durability :)
			durability = GetItemDurability(s);
			if( not durability ) then
				s[SLOT_CAN_DURABILITY] = false;
			else
				s[SLOT_DURABILITY] = durability;
				
				if( durability < smallest_durability ) then
					smallest_durability = durability;
				end
				
				RepairCosts = RepairCosts + repcosts;
			end
			
			-- gems
			s[SLOT_GEMS_EMPTY] = NumMissingGems(s[SLOT_ID]);
			-- retrieve item color, name and enchant
			s[SLOT_COLOR], enchant, s[SLOT_ITEM_NAME] = GetItemInfos(s[SLOT_ID]);
			-- enchant
			s[SLOT_ENCHANT] = tonumber(enchant);
		end
	end
	
	self.Feed.text = ("|cff%s%d%%|r"):format(LibCrayon:GetThresholdHexColor(smallest_durability, 100), smallest_durability);
	
	local problems = self:GetNumProblems();
	if( problems and problems ~= 0 ) then
		self.Feed.text = ("%s |cffff0000%d!|r"):format(self.Feed.text, problems);
	end
end

-----------------------
-- UpdateTooltip
-----------------------

local function GetHintText(slot)
	local text = "";
	
	if( slot[SLOT_GEMS_EMPTY] ~= 0 ) then
		text = ("|cff%s%d %s|r"):format("ff0000", slot[SLOT_GEMS_EMPTY], slot[SLOT_GEMS_EMPTY] == 1 and L["Gem"] or L["Gems"]);
	end
	
	if( slot[SLOT_CAN_ENCHANT] and slot[SLOT_ENCHANT] == 0 ) then
		text = ("%s%s|cff00ffff%s|r"):format(text, text ~= "" and ", " or "", L["Enchant"]);
	end
	
	return text;
end

local function LineEnter(anchor, slotNum)
	_G.GameTooltip_SetDefaultAnchor(_G.GameTooltip, _G.UIParent);
	_G.GameTooltip:SetInventoryItem("player", slotNum);
	_G.GameTooltip:Show();
end

local function LineLeave()
	_G.GameTooltip:ClearLines();
	_G.GameTooltip:Hide();
end

local TipAlign= {};
local TipHead = {};
local TipLine = {};
function iGear:UpdateTooltip()
	Tooltip:Clear();
	
	-- cols: durability, slot, icon, item, hints
	tclear(TipAlign);
	tclear(TipHead);
	
	-- we just show the durability column if there's something to repair (beautiful!)
	if( RepairCosts ~= 0 ) then
		table.insert(TipAlign, "RIGHT");
		table.insert(TipHead, "");
	end
	
	table.insert(TipAlign, "LEFT");
	table.insert(TipAlign, "RIGHT");
	table.insert(TipAlign, "LEFT");
	
	table.insert(TipHead, L["Slot"]);
	table.insert(TipHead, "");
	table.insert(TipHead, L["Item"]);
	
	local problems = self:GetNumProblems();
	if( problems > 0 ) then
		table.insert(TipAlign, "LEFT");
		table.insert(TipHead, L["Hints"]);
	end
	
	Tooltip:SetColumnLayout(#TipAlign, unpack(TipAlign)); -- I really love the unpack function! ;)
	Tooltip:AddHeader(unpack(TipHead)); -- here as well!
	
	local line;
	for _, s in ipairs(EquipSlots) do
		if( self:GetProblem(s, 1) ) then
			tclear(TipLine);
			
			if( RepairCosts ~= 0 ) then
				if( s[SLOT_CAN_DURABILITY] and s[SLOT_DURABILITY] ~= 100 ) then
					table.insert(TipLine, ("|cff%s%d%%|r"):format(LibCrayon:GetThresholdHexColor(s[SLOT_DURABILITY], 100), s[SLOT_DURABILITY]));
				else
					table.insert(TipLine, "");
				end
			end
			
			table.insert(TipLine, (COLOR_GOLD):format(L[s[SLOT_NAME]]));
			table.insert(TipLine, ("|T%s:14:14|t"):format(_G.GetInventoryItemTexture("player", s[SLOT_ID])));
			table.insert(TipLine, ("|cff%s%s|r"):format(s[SLOT_COLOR], s[SLOT_ITEM_NAME] ~= "" and s[SLOT_ITEM_NAME] or UNKNOWN));
			
			if( problems > 0 ) then
				table.insert(TipLine, GetHintText(s));
			end
			
			line = Tooltip:AddLine(unpack(TipLine));
			Tooltip:SetLineScript(line, "OnEnter", LineEnter, s[SLOT_ID]);
			Tooltip:SetLineScript(line, "OnLeave", LineLeave);
		end
	end
	
	if( RepairCosts ~= 0 ) then
		Tooltip:AddLine(" ");
		Tooltip:AddLine("", (COLOR_GOLD):format(L["Repair Costs"]), "", _G.GetMoneyString(RepairCosts));
		Tooltip:AddLine("", (COLOR_GOLD):format(FACTION_STANDING_LABEL8), "", _G.GetMoneyString(RepairCosts *0.8));
	end
end